"""
View Stellarcarbon's current or historical inventory of eco-credits.

The inventory is calculated as: all blocks of credits that have been minted, minus the retirements
that have been taken from these blocks. Historical views reconstruct the inventory on the given
date (end-of-day) by filtering both the blocks and the retirements used in the calculation.

Author: Alex Olieman <https://keybase.io/alioli>
"""
import datetime as dt

import pandas as pd
from sqlalchemy import select

from sc_audit.db_schema.impact_project import VcsProject
from sc_audit.db_schema.mint import MintedBlock
from sc_audit.session_manager import Session


def view_inventory(omit_empty: bool = False, until_date: dt.date | None = None) -> pd.DataFrame:
    credits_remaining_col = MintedBlock.credits_remaining
    if until_date:
        credits_remaining_col = MintedBlock.credits_remaining_on_date(until_date)

    block_columns = [
        MintedBlock.serial_hash,
        MintedBlock.size,
        credits_remaining_col,
        MintedBlock.transaction_hash,
        MintedBlock.created_at,
        MintedBlock.serial_number,
        MintedBlock.sub_account_id,
        MintedBlock.sub_account_name,
        MintedBlock.vintage_start,
        MintedBlock.vintage_end,
    ]
    vcs_project_columns = [
        MintedBlock.vcs_project_id,
        VcsProject.name,
        VcsProject.category,
    ]    
    columns = block_columns + vcs_project_columns
    q_blocks = select(*columns).join(MintedBlock.vcs_project).order_by(MintedBlock.created_at)
    if omit_empty:
        q_blocks = q_blocks.where(credits_remaining_col > 0)

    if until_date:
        until_dt = dt.datetime(until_date.year, until_date.month, until_date.day, 23, 59, 59, 999999)
        q_blocks = q_blocks.where(MintedBlock.created_at <= until_dt)

    with Session.begin() as session:
        mb_rows = session.execute(q_blocks).all()
        mbdf = pd.DataFrame.from_records(mb_rows, columns=(c.key for c in columns))

    return mbdf



"""
Load finalized retirements into the DB.

Author: Alex Olieman <https://keybase.io/alioli>
"""
import datetime as dt
from functools import partial

from sqlalchemy import select

from sc_audit.db_schema.impact_project import VcsProject
from sc_audit.db_schema.retirement import Retirement
from sc_audit.loader.utils import parse_iso_date
from sc_audit.session_manager import Session
from sc_audit.sources.retirements import get_retirements_list


def load_retirements(from_date: dt.date | None = None) -> int:
    """
    Load all retirements from Verra into the DB.
  
    TODO: add pagination (to support more than 2000 results)
    """
    retirement_data = get_retirements_list(from_date)
    number_loaded = 0

    with Session.begin() as session:
        existing_ids: set[int] = set(session.scalars(select(Retirement.certificate_id)).all())
        existing_vcs_projects: set[int] = set(session.scalars(select(VcsProject.id)).all())
        for retirement_item in retirement_data['retirements']:
            # only load new retirements
            certificate_id = int(retirement_item['certificate_id'])
            if certificate_id not in existing_ids:
                # ensure that the related VCS Project exists
                vcs_project_id = int(retirement_item['vcs_id'])
                if vcs_project_id not in existing_vcs_projects:
                    existing_vcs_projects.add(vcs_project_id)
                    session.add(
                        VcsProject(
                            id=vcs_project_id,
                            name=retirement_item['vcs_name'],
                            category=retirement_item['vcs_category'],
                            protocol=retirement_item['vcs_protocol'],
                            additional_certifications=retirement_item['additional_certifications'],
                            region=retirement_item['region'],
                            country=retirement_item['country'],
                        )
                    )

                session.add(
                    Retirement(
                        certificate_id=certificate_id,
                        vcu_amount=retirement_item['vcu_amount'],
                        serial_number=retirement_item['serial_numbers'],
                        retirement_date=parse_iso_date(retirement_item['retirement_date']),
                        retirement_beneficiary=retirement_item['retirement_beneficiary'],
                        retirement_details=retirement_item['retirement_details'],
                        vcs_project_id=vcs_project_id,
                        issuance_date=parse_iso_date(retirement_item['issuance_date']),
                        instrument_type=retirement_item['instrument_type'],
                        vintage_start=parse_iso_date(retirement_item['vintage_start']),
                        vintage_end=parse_iso_date(retirement_item['vintage_end']),
                        total_vintage_quantity=retirement_item['total_vintage_quantity'],
                    )
                )
                number_loaded += 1

    return number_loaded



"""
Dump DB tables to newline-delimited json files.

Author: Alex Olieman <https://keybase.io/alioli>
"""
import inspect
from pathlib import Path

import pandas as pd
from sqlalchemy import select, Table

from sc_audit.db_schema.base import ScBase
from sc_audit.session_manager import Session


def dump_table(db_model: type[ScBase] | str, output_path: Path | None = None) -> str | None:
    db_table: Table
    if inspect.isclass(db_model) and issubclass(db_model, ScBase):
        db_table = db_model.__table__ # type: ignore
    else:
        try:
            db_table = ScBase.metadata.tables[db_model]
        except KeyError:
            table_names = get_table_names()
            raise KeyError(f"`db_model` must be one of {table_names}")
        
    ndjson_table = table_to_ndjson(db_table)
    if output_path is not None:
        with output_path.open('w') as ofile:
            ofile.write(ndjson_table)
    else:
        return ndjson_table


def table_to_ndjson(db_table: Table) -> str:
    dbq = select(db_table)
    columns = db_table.columns
    with Session.begin() as session:
        rows = session.execute(dbq).all()
        tdf = pd.DataFrame.from_records(rows, columns=(c.name for c in columns))

    return tdf.to_json(
        orient='values',
        date_format='iso',
        date_unit='s',
    )[1:-1].replace("],[", "]\n[")


def get_table_names() -> list[str]:
    return [
        table_name
        for table_name in ScBase.metadata.tables.keys()
        if not table_name.startswith("test_")
    ]



from __future__ import annotations
import base64
import datetime

from pydantic import BaseModel

def parse_iso_date(iso_date: str) -> datetime.date:
    return datetime.datetime.fromisoformat(iso_date).date()

def parse_iso_datetime(iso_datetime: str) -> datetime.datetime:
    return datetime.datetime.fromisoformat(iso_datetime)

def decode_hash_memo(b64_hash_memo: str) -> str:
    return base64.b64decode(b64_hash_memo).hex()


class VcsSerialNumber(BaseModel):
    batch_number: int
    block_start: int
    block_end: int
    program_code: str | None
    credit_type: str
    vvb_id: str
    registry: str
    country: str
    project_type: int
    project_id: int
    vintage_start: str
    vintage_end: str
    additional_certification: bool

    @classmethod
    def from_str(cls, vcs_serial_number: str) -> VcsSerialNumber:
        split_serial = vcs_serial_number.split('-')
        if len(split_serial) == 12:
            split_serial = [*split_serial[:3], None, *split_serial[3:]]
        
        kwargs = dict(zip(VcsSerialNumber.model_fields, split_serial))
        return VcsSerialNumber(**kwargs) # type: ignore
    
    def to_str(self) -> str:
        serial_dict = self.model_dump()
        if self.program_code is None:
            del serial_dict['program_code']

        serial_dict['additional_certification'] = int(self.additional_certification)
        return '-'.join(map(str, serial_dict.values()))
    
    @property
    def vintage_start_date(self) -> datetime.date:
        return datetime.datetime.strptime(self.vintage_start, "%d%m%Y").date()
    
    @property
    def vintage_end_date(self) -> datetime.date:
        return datetime.datetime.strptime(self.vintage_end, "%d%m%Y").date()



from __future__ import annotations
import typing

from sqlalchemy import ForeignKey, Index
from sqlalchemy.orm import Mapped, MappedAsDataclass, mapped_column, relationship

from sc_audit.db_schema.base import ScBase, intpk, hashpk, kgdecimal

if typing.TYPE_CHECKING:
    from sc_audit.db_schema import MintedBlock, Retirement, SinkingTx


class RetirementFromBlockBase(MappedAsDataclass, kw_only=True):

    retirement_id: Mapped[intpk] = mapped_column(ForeignKey('retirements.certificate_id'))
    block_hash: Mapped[hashpk] = mapped_column(ForeignKey('minted_blocks.serial_hash'))
    vcu_amount: Mapped[int]


class RetirementFromBlock(RetirementFromBlockBase, ScBase):
    __tablename__ = "retirement_from_block"

    retirement: Mapped[Retirement] = relationship(
        init=False, repr=False,
        back_populates='retired_from'
    )
    block: Mapped[MintedBlock] = relationship(
        init=False, repr=False, 
        back_populates='consumed_by'
    )

    def as_dict(self):
        return {
            col: getattr(self, col)
            for col in self.__table__.columns.keys()
        }


class SinkStatusBase(MappedAsDataclass, kw_only=True):

    sinking_tx_hash: Mapped[hashpk] = mapped_column(ForeignKey('sinking_txs.hash'))
    certificate_id: Mapped[intpk] = mapped_column(ForeignKey('retirements.certificate_id'))
    amount_filled: Mapped[kgdecimal]
    finalized: Mapped[bool]


class SinkStatus(SinkStatusBase, ScBase):
    __tablename__ = "sink_status"

    sinking_transaction: Mapped[SinkingTx] = relationship(
        init=False, repr=False, 
        back_populates='statuses'
    )
    retirement: Mapped[Retirement] = relationship(
        init=False, repr=False,
        back_populates='sink_statuses'
    )

    def as_dict(self):
        return {
            col: getattr(self, col)
            for col in self.__table__.columns.keys()
        }
    

idx_finalized = Index("idx_finalized", SinkStatus.finalized)



"""add test prefix tables

Revision ID: 19f853853773
Revises: 0e7b244f69dc
Create Date: 2024-07-25 19:00:55.917243

"""
from typing import Sequence

from alembic import op
import sqlalchemy as sa
from sc_audit.db_schema.base import HexBinary

# revision identifiers, used by Alembic.
revision: str = '19f853853773'
down_revision: str | None = '0e7b244f69dc'
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    op.create_table('test_distribution_txs',
    sa.Column('hash', HexBinary(length=32), nullable=False),
    sa.Column('created_at', sa.DateTime(), nullable=False),
    sa.Column('sender', sa.String(length=56), nullable=False),
    sa.Column('recipient', sa.String(length=56), nullable=False),
    sa.Column('carbon_amount', sa.DECIMAL(precision=21, scale=3), nullable=False),
    sa.Column('paging_token', sa.Integer(), nullable=False),
    sa.PrimaryKeyConstraint('hash')
    )
    op.create_table('test_vcs_projects',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(), nullable=False),
    sa.Column('category', sa.Enum('Agriculture Forestry and Other Land Use', native_enum=False), nullable=False),
    sa.Column('protocol', sa.Enum('VM0015', native_enum=False), nullable=False),
    sa.Column('additional_certifications', sa.String(), nullable=True),
    sa.Column('region', sa.Unicode(length=128), nullable=False),
    sa.Column('country', sa.Unicode(length=128), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('test_minted_blocks',
    sa.Column('serial_hash', HexBinary(length=32), nullable=False),
    sa.Column('transaction_hash', HexBinary(length=32), nullable=False),
    sa.Column('created_at', sa.DateTime(), nullable=False),
    sa.Column('vcs_project_id', sa.Integer(), nullable=False),
    sa.Column('serial_number', sa.String(length=128), nullable=False),
    sa.Column('block_start', sa.Integer(), nullable=False),
    sa.Column('block_end', sa.Integer(), nullable=False),
    sa.Column('sub_account_id', sa.Integer(), nullable=False),
    sa.Column('sub_account_name', sa.Enum('CARBON Pool | stellarcarbon.io', 'CARBON Sink | stellarcarbon.io', native_enum=False), nullable=False),
    sa.Column('vintage_start', sa.Date(), nullable=False),
    sa.Column('vintage_end', sa.Date(), nullable=False),
    sa.Column('paging_token', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['vcs_project_id'], ['vcs_projects.id'], ),
    sa.PrimaryKeyConstraint('serial_hash')
    )
    op.create_table('test_retirements',
    sa.Column('certificate_id', sa.Integer(), nullable=False),
    sa.Column('vcu_amount', sa.Integer(), nullable=False),
    sa.Column('serial_number', sa.String(length=128), nullable=False),
    sa.Column('retirement_date', sa.Date(), nullable=False),
    sa.Column('retirement_beneficiary', sa.String(length=56), nullable=False),
    sa.Column('retirement_details', sa.String(), nullable=False),
    sa.Column('vcs_project_id', sa.Integer(), nullable=False),
    sa.Column('issuance_date', sa.Date(), nullable=False),
    sa.Column('instrument_type', sa.Enum('VCU', native_enum=False), nullable=False),
    sa.Column('vintage_start', sa.Date(), nullable=False),
    sa.Column('vintage_end', sa.Date(), nullable=False),
    sa.Column('total_vintage_quantity', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['vcs_project_id'], ['vcs_projects.id'], ),
    sa.PrimaryKeyConstraint('certificate_id')
    )
    op.create_table('test_sinking_txs',
    sa.Column('hash', HexBinary(length=32), nullable=False),
    sa.Column('created_at', sa.DateTime(), nullable=False),
    sa.Column('funder', sa.String(length=56), nullable=False),
    sa.Column('recipient', sa.String(length=56), nullable=False),
    sa.Column('carbon_amount', sa.DECIMAL(precision=21, scale=3), nullable=False),
    sa.Column('source_asset_code', sa.String(length=12), nullable=False),
    sa.Column('source_asset_issuer', sa.String(length=56), nullable=True),
    sa.Column('source_asset_amount', sa.DECIMAL(precision=21, scale=7), nullable=False),
    sa.Column('dest_asset_code', sa.String(length=12), nullable=False),
    sa.Column('dest_asset_issuer', sa.String(length=56), nullable=False),
    sa.Column('dest_asset_amount', sa.DECIMAL(precision=21, scale=7), nullable=False),
    sa.Column('vcs_project_id', sa.Integer(), nullable=False),
    sa.Column('memo_type', sa.Enum('text', 'hash', 'none', native_enum=False), nullable=False),
    sa.Column('memo_value', sa.String(length=64), nullable=True),
    sa.Column('paging_token', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['vcs_project_id'], ['vcs_projects.id'], ),
    sa.PrimaryKeyConstraint('hash')
    )
    op.create_table('test_retirement_from_block',
    sa.Column('retirement_id', sa.Integer(), nullable=False),
    sa.Column('block_hash', HexBinary(length=32), nullable=False),
    sa.Column('vcu_amount', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['block_hash'], ['minted_blocks.serial_hash'], ),
    sa.ForeignKeyConstraint(['retirement_id'], ['retirements.certificate_id'], ),
    sa.PrimaryKeyConstraint('retirement_id', 'block_hash')
    )
    op.create_table('test_sink_status',
    sa.Column('sinking_tx_hash', HexBinary(length=32), nullable=False),
    sa.Column('certificate_id', sa.Integer(), nullable=False),
    sa.Column('amount_filled', sa.DECIMAL(precision=21, scale=3), nullable=False),
    sa.Column('finalized', sa.Boolean(), nullable=False),
    sa.ForeignKeyConstraint(['certificate_id'], ['retirements.certificate_id'], ),
    sa.ForeignKeyConstraint(['sinking_tx_hash'], ['sinking_txs.hash'], ),
    sa.PrimaryKeyConstraint('sinking_tx_hash', 'certificate_id')
    )


def downgrade() -> None:
    op.drop_table('test_sink_status')
    op.drop_table('test_retirement_from_block')
    op.drop_table('test_sinking_txs')
    op.drop_table('test_retirements')
    op.drop_table('test_minted_blocks')
    op.drop_table('test_vcs_projects')
    op.drop_table('test_distribution_txs')



"""
Load minted blocks of CARBON into the DB.

Minted blocks are reconstructed using the relationship between minting transactions and
registry sources. Every carbon credit that is minted is either still in the inventory
or it has already been retired. Blocks that have not yet been used for retirements have a
corresponding serial number hash memo on their minting tx. Serial numbers of any blocks that
have been fully or partly used for retirements are reconstructed on the basis of the serial
numbers of the retired credits.

Author: Alex Olieman <https://keybase.io/alioli>
"""
from hashlib import sha256
from typing import Any, Literal, Sequence

from sqlalchemy import select

from sc_audit.config import settings
from sc_audit.db_schema.impact_project import UnknownVcsProject, VcsProject, get_vcs_project
from sc_audit.db_schema.mint import MintedBlock, verra_carbon_pool
from sc_audit.db_schema.retirement import Retirement
from sc_audit.loader.utils import VcsSerialNumber, decode_hash_memo, parse_iso_datetime
from sc_audit.session_manager import Session
from sc_audit.sources.carbon_pool import get_carbon_pool_state
from sc_audit.sources.minting_txs import get_minting_transactions


def load_minted_blocks(cursor: int=settings.FIRST_MINT_CURSOR) -> int:
    """
    Load (all) minting transactions from Horizon into the DB.

    To catch up with Horizon, specify the cursor parameter to be the incremented paging
    token of the latest record present in the DB.
    """
    retired_blocks: list[dict[str, Any]] = []
    pristine_blocks: list[MintedBlock] = []
    carbon_pool = index_carbon_pool()
    with Session.begin() as session:
        existing_vcs_projects: set[int] = set(session.scalars(select(VcsProject.id)).all())
        for mint_tx in get_minting_transactions(cursor):
            # ensure that the related VCS Project exists
            vcs_project_id = get_vcs_project_id(mint_tx)
            if vcs_project_id not in existing_vcs_projects:
                vcs_project = get_vcs_project(vcs_project_id)
                if vcs_project:
                    existing_vcs_projects.add(vcs_project_id)
                    session.add(vcs_project)
                else:
                    raise UnknownVcsProject(
                        f"VCS project {vcs_project_id} needs to be loaded before related transactions"
                        " can be stored.",
                        vcs_id=vcs_project_id
                    )
            # check whether the block is pristine
            serial_hash = decode_hash_memo(mint_tx['transaction']['memo'])
            if serial_hash in carbon_pool:
                serial_number = VcsSerialNumber.from_str(carbon_pool[serial_hash]['serial_number']) # type: ignore
                block = MintedBlock(
                    serial_hash=serial_hash,
                    transaction_hash=mint_tx['transaction_hash'],
                    created_at=parse_iso_datetime(mint_tx['created_at']),
                    vcs_project_id=vcs_project_id,
                    serial_number=serial_number.to_str(),
                    block_start=serial_number.block_start,
                    block_end=serial_number.block_end,
                    sub_account_id=carbon_pool[serial_hash]['sub_account_id'], # type: ignore
                    sub_account_name=carbon_pool[serial_hash]['sub_account_name'], # type: ignore
                    vintage_start=serial_number.vintage_start_date,
                    vintage_end=serial_number.vintage_end_date,
                    paging_token=mint_tx['paging_token']
                )
                pristine_blocks.append(block)
            else:
                retired_blocks.append(mint_tx)

        if retired_blocks:
            latest_loaded_block: MintedBlock | None = session.scalars(
                select(MintedBlock).order_by(MintedBlock.paging_token.desc())
            ).first()
            # select only retirements without retired_from_block relations
            retirements: Sequence[Retirement] = session.scalars(
                select(Retirement).where(~Retirement.retired_from.any())
            ).all()
            reconstructed_blocks = reconstruct_blocks(
                retired_blocks, latest_loaded_block, retirements
            )
            session.add_all(reconstructed_blocks)

        session.add_all(pristine_blocks)

    return len(retired_blocks) + len(pristine_blocks)
            
            
def get_vcs_project_id(sinking_tx) -> Literal[1360]:
    # TODO: support multiple VCS projects
    return 1360

def index_carbon_pool() -> dict[str, dict[str, str | int]]:
    """
    Index current blocks in the carbon pool by sha256(serial_number).    
    """
    indexed_pool = {}
    carbon_pool = get_carbon_pool_state()
    for block in carbon_pool['credit_batches']:
        serial_number_hash = sha256(block['serial_number'].encode('utf8')).hexdigest()
        indexed_pool[serial_number_hash] = block

    return indexed_pool

def reconstruct_blocks(
        mint_txs: list[dict[str, Any]],
        latest_block: MintedBlock | None,
        retirements: Sequence[Retirement]
    ) -> list[MintedBlock]:
    """
    Attempt to reconstruct blocks by:
    1. checking if the block directly follows the last known block
    2. finding the first retirement done from the block
    """
    reconstructed_blocks: list[MintedBlock] = []
    for mint_tx in mint_txs:
        vcs_project_id = get_vcs_project_id(mint_tx)
        serial_hash = decode_hash_memo(mint_tx['transaction']['memo'])
        block_size = int(float(mint_tx['amount']))
        # check if this block is an extension of the previous block
        if latest_block:
            reconstructed_serial = VcsSerialNumber.from_str(latest_block.serial_number)
            reconstructed_serial.block_start = 1 + reconstructed_serial.block_end
            reconstructed_serial.block_end = (block_size - 1) + reconstructed_serial.block_start
            sub_account_id = latest_block.sub_account_id
            sub_account_name = latest_block.sub_account_name

        # keep looking, unless there is a latest block and the reconstructed serial matches the known hash
        if not (latest_block and serial_matches_hash(reconstructed_serial, serial_hash)):
            for retirement in retirements:
                # find the retirement of the first credit from this block
                reconstructed_serial = VcsSerialNumber.from_str(retirement.serial_number)
                reconstructed_serial.block_end = (block_size - 1) + reconstructed_serial.block_start
                if serial_matches_hash(reconstructed_serial, serial_hash):
                    sub_account_id = verra_carbon_pool.id
                    sub_account_name = verra_carbon_pool.name
                    break
            else:
                raise ValueError(f"No match found for serial hash {serial_hash}.", serial_hash)

        block = MintedBlock(
            serial_hash=serial_hash,
            transaction_hash=mint_tx['transaction_hash'],
            created_at=parse_iso_datetime(mint_tx['created_at']),
            vcs_project_id=vcs_project_id,
            serial_number=reconstructed_serial.to_str(),
            block_start=reconstructed_serial.block_start,
            block_end=reconstructed_serial.block_end,
            sub_account_id=sub_account_id,
            sub_account_name=sub_account_name,
            vintage_start=reconstructed_serial.vintage_start_date,
            vintage_end=reconstructed_serial.vintage_end_date,
            paging_token=mint_tx['paging_token']
        )
        latest_block = block
        reconstructed_blocks.append(block)

    assert len(reconstructed_blocks) == len(mint_txs)
    return reconstructed_blocks

def serial_matches_hash(reconstructed_serial: VcsSerialNumber, serial_hash: str) -> bool:
    """
    Test whether the given serial number corresponds to the known hash.
    """
    reconstructed_serial_hash = sha256(
        reconstructed_serial.to_str().encode('utf8')
    ).hexdigest()
    return serial_hash == reconstructed_serial_hash



from pathlib import Path
from typing import Literal

from pydantic import HttpUrl, computed_field
from pydantic_core import Url
from pydantic_settings import BaseSettings, SettingsConfigDict
from stellar_sdk import Asset


def get_default_db_path(db_name="sc-audit.sqlite3") -> Path:
    """
    Return the default location of the SQLite DB, as determined by:

    if the current working dir is a descendent of the home dir:
        use the current working dir
    else:
        use a subdir of the home dir and ensure it exists
    """
    home_dir = Path.home()
    work_dir = Path.cwd()

    if home_dir in work_dir.resolve().parents:
        if work_dir.name == "sc_audit":
            return work_dir.parent / db_name
        
        return work_dir / db_name
    else:
        db_dir = home_dir / "sc-audit"
        db_dir.mkdir(exist_ok=True)
        return db_dir / db_name
    

DbTablePrefix = Literal["test"] | None


class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix='SC_')

    DBAPI_URL: str = f"sqlite+pysqlite:///{get_default_db_path()}"
    TABLE_PREFIX: DbTablePrefix = None

    VERRA_ASSET_SEARCH_URL: HttpUrl = Url("https://registry.verra.org/uiapi/asset/asset/search")
    VERRA_ASSET_SEARCH_TIMEOUT: int = 8  # seconds
    VERRA_REPORT_URL: HttpUrl = Url("https://registry.verra.org/mymodule/rpt/myRpt.asp")

    HORIZON_URL: HttpUrl = Url("https://horizon.stellar.org")
    CARBON_ISSUER_PUB: str = "GCBOATLWKXACOWKRRWORARDI2HFDSYPALMTS23YBZKHOB6XLW6CARBON"
    CARBON_DISTRIB_PUB: str = "GABXJLJDNWRIGCVQPPC2FJ2NORUJBGLNOBYXKW7WQD3CUZS6YTFCS4ZP"
    SINK_ISSUER_PUB: str = "GC7CDWMCWNCY7JYUW5UBEOLNBSTNDKKZSFTHKGZNPPSOXLFYFX3CSINK"
    FIRST_SINK_CURSOR: int = 164821723627237376
    FIRST_MINT_CURSOR: int = 164806777139924992
    FIRST_DIST_CURSOR: int = 164810659791396865

    @computed_field
    @property
    def CARBON_ASSET(self) -> Asset:
        return Asset("CARBON", self.CARBON_ISSUER_PUB)

    @computed_field
    @property
    def SINK_ASSET(self) -> Asset:
        return Asset("CarbonSINK", self.SINK_ISSUER_PUB)


settings = Settings()



"""
View retirements and their details.

Retirements can be filtered by beneficiary, by retirement date, and by impact project.

Author: Alex Olieman <https://keybase.io/alioli>
"""
import datetime as dt
from typing import Any, Literal

import pandas as pd
from sqlalchemy import select
from sqlalchemy.orm import selectinload

from sc_audit.db_schema.retirement import Retirement
from sc_audit.session_manager import Session


def view_retirements(
        for_beneficiary: str | None = None, 
        from_date: dt.date | None = None,
        before_date: dt.date | None = None,
        project: int | None = None,
        cursor: int | None = None,
        limit: int | None = None,
        order: Literal['asc', 'desc'] = 'desc',
) -> pd.DataFrame:
    rt_query = construct_retirement_query(for_beneficiary, from_date, before_date, project)

    if order == 'asc':
        rt_query = rt_query.order_by(Retirement.certificate_id.asc())
        if cursor:
            rt_query = rt_query.where(Retirement.certificate_id > cursor)
    if order == 'desc':
        rt_query = rt_query.order_by(Retirement.certificate_id.desc())
        if cursor:
            rt_query = rt_query.where(Retirement.certificate_id < cursor)

    if limit:
        rt_query = rt_query.limit(limit)

    with Session.begin() as session:
        rt_records = session.scalars(rt_query).all()
        rtdf = pd.DataFrame.from_records(ret.as_dict() for ret in rt_records)

    return rtdf


def construct_retirement_query(
        for_beneficiary: str | None, 
        from_date: dt.date | None,
        before_date: dt.date | None,
        project: int | None,
):
    q_rets = select(Retirement)
    if for_beneficiary:
        q_rets = q_rets.where(Retirement.retirement_beneficiary.startswith(for_beneficiary))

    if from_date:
        q_rets = q_rets.where(Retirement.retirement_date >= from_date)

    if before_date:
        q_rets = q_rets.where(Retirement.retirement_date < before_date)

    if project:
        q_rets = q_rets.where(Retirement.vcs_project_id == project)

    return q_rets


def get_retirement(certificate_id: int) -> dict[str, Any] | None:
    rt_query = (
        select(Retirement)
        .options(
            selectinload(Retirement.retired_from), 
            selectinload(Retirement.sink_statuses),
            selectinload(Retirement.vcs_project)
        )
        .where(Retirement.certificate_id == certificate_id)
    )
    with Session.begin() as session:
        ret = session.scalar(rt_query)
        ret_data = ret.as_dict(related=True) if ret else None

    return ret_data



"""add distribution outflows table

Revision ID: 0e7b244f69dc
Revises: b595c6a10724
Create Date: 2024-06-24 17:18:49.287452

"""
from typing import Sequence

from alembic import op
import sqlalchemy as sa
from sc_audit.db_schema.base import HexBinary

# revision identifiers, used by Alembic.
revision: str = '0e7b244f69dc'
down_revision: str | None = 'b595c6a10724'
branch_labels: str | Sequence[str] | None = None
depends_on: str | Sequence[str] | None = None


def upgrade() -> None:
    op.create_table('distribution_txs',
    sa.Column('hash', HexBinary(length=32), nullable=False),
    sa.Column('created_at', sa.DateTime(), nullable=False),
    sa.Column('sender', sa.String(length=56), nullable=False),
    sa.Column('recipient', sa.String(length=56), nullable=False),
    sa.Column('carbon_amount', sa.DECIMAL(precision=21, scale=3), nullable=False),
    sa.Column('paging_token', sa.Integer(), nullable=False),
    sa.PrimaryKeyConstraint('hash')
    )
    op.create_index('idx_dtx_created_at', 'distribution_txs', [sa.text('created_at DESC')], unique=False)
    op.create_index('idx_dtx_recipient', 'distribution_txs', ['recipient'], unique=False)


def downgrade() -> None:
    op.drop_index('idx_dtx_recipient', table_name='distribution_txs')
    op.drop_index('idx_dtx_created_at', table_name='distribution_txs')
    op.drop_table('distribution_txs')



"""
Restore DB tables from newline-delimited json files.

Author: Alex Olieman <https://keybase.io/alioli>
"""
from io import StringIO
from pathlib import Path
import sys

import pandas as pd
from sqlalchemy import delete, func, select, types

from sc_audit.db_schema import (
    ScBase, DistributionTx, MintedBlock, Retirement, RetirementFromBlock, SinkingTx, SinkStatus, VcsProject
)
from sc_audit.session_manager import Session


TABLE_LOADING_ORDER = (
    VcsProject,
    DistributionTx,
    Retirement,
    SinkingTx,
    MintedBlock,
    RetirementFromBlock,
    SinkStatus,
)

def restore_all_tables(input_dir: Path, replace: bool = False):
    for db_model in TABLE_LOADING_ORDER:
        table_name = db_model.__table__.name # type: ignore
        dump_file = input_dir / f"{table_name}.ndjson"
        if dump_file.exists():
            restore_table(dump_file, replace=replace)
        else:
            print(f"Couldn't find {dump_file}", file=sys.stderr)


def restore_table(json_path: Path, replace: bool = False):
    table_name = json_path.stem
    db_table = ScBase.metadata.tables[table_name]
    tdf = df_from_ndjson(json_path)

    with Session.begin() as session:
        if replace:
            # delete existing rows
            dbres = session.execute(delete(db_table))
            print(f"Deleted {dbres.rowcount} existing rows from {table_name}")
        else:
            # exit early if there are existing rows
            num_rows = session.scalar(select(func.count()).select_from(db_table))
            if num_rows:
                print(f"Leaving {num_rows} rows from {table_name} intact")
                return
            
        # insert the dataframe records into the table
        column_dtypes = {c.name: c.type for c in db_table.columns}
        num_rows = tdf.to_sql(
            name=table_name,
            con=session.connection(),
            if_exists='append',
            index=False,
            chunksize=200,
            dtype=column_dtypes, # type: ignore
        )
        print(f"Restored {num_rows} rows into {table_name}")


def df_from_ndjson(json_path: Path) -> pd.DataFrame:
    table_name = json_path.stem
    db_table = ScBase.metadata.tables[table_name]
    with open(json_path, 'r') as infile:
        ndjson = infile.read()

    # turn json lines into a valid array
    json_array = "[" + ndjson.replace("\n", ",") + "]"
    table_df = pd.read_json(
        StringIO(json_array),
        orient='values',
    )
    column_dtypes = {c.name: c.type for c in db_table.columns}
    table_df.columns = list(column_dtypes.keys())

    # parse date(time) strings
    for col_name, col_type in column_dtypes.items():
        if isinstance(col_type, (types.Date, types.DateTime)):
            table_df[col_name] = pd.to_datetime(table_df[col_name], format='ISO8601')

    return table_df


def get_table_row_counts() -> dict[str, int]:
    row_counts: dict[str, int] = {}
    with Session.begin() as session:
        for table_name, table in ScBase.metadata.tables.items():
            num_rows = session.scalar(select(func.count()).select_from(table))
            row_counts[table_name] = num_rows or 0

    return row_counts



"""
Fetch sinking transactions from Stellar Horizon.

Author: Alex Olieman <https://keybase.io/alioli>
"""
from stellar_sdk import Server
from stellar_sdk.server import OperationsCallBuilder, PaymentsCallBuilder

from sc_audit.config import settings


server = Server(horizon_url=str(settings.HORIZON_URL))


def get_sinking_transactions(cursor: int=settings.FIRST_SINK_CURSOR):
    query: PaymentsCallBuilder = (
        server.payments()
        .for_account(settings.SINK_ISSUER_PUB)
        .cursor(cursor)
        .join("transactions")
        .order(desc=False)
        .limit(200)
    )
    resp = query.call()
    yield from filter_sinking_txs(resp['_embedded']['records'])
    while next_records := query.next()['_embedded']['records']:
        yield from filter_sinking_txs(next_records)

def filter_sinking_txs(horizon_records):
    for payment in horizon_records:
        if (
            payment["type_i"] == 1
            and payment['from'] == settings.SINK_ISSUER_PUB
            and payment['asset_code'] == settings.SINK_ASSET.code
        ):
            yield payment

def get_tx_operations(tx_hash: str) -> list[dict]:
    query: OperationsCallBuilder = server.operations().for_transaction(tx_hash)
    resp = query.call()
    return resp['_embedded']['records']



"""
Declare CARBON distribution DB models.

Author: Alex Olieman <https://keybase.io/alioli>
"""
import datetime as dt

from sqlalchemy import Index
from sqlalchemy.orm import Mapped, MappedAsDataclass

from sc_audit.db_schema.base import ScBase, bigint, hashpk, kgdecimal, strkey


class DistributionTxBase(MappedAsDataclass):

    hash: Mapped[hashpk]
    created_at: Mapped[dt.datetime]
    sender: Mapped[strkey]
    recipient: Mapped[strkey]
    carbon_amount: Mapped[kgdecimal]
    paging_token: Mapped[bigint]


class DistributionTx(DistributionTxBase, ScBase):
    __tablename__ = "distribution_txs"


idx_created_at = Index("idx_dtx_created_at", DistributionTx.created_at.desc())
idx_recipient = Index("idx_dtx_recipient", DistributionTx.recipient)



"""
DB connection management.

Author: Alex Olieman <https://keybase.io/alioli>
"""
import sys
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from sc_audit.config import settings
from sc_audit.db_schema.base import create_test_mappers

print(f"Connecting to database {settings.DBAPI_URL}...", file=sys.stderr)
engine = create_engine(settings.DBAPI_URL)
Session = sessionmaker(engine)

# create test mappers after all models have been registered
# TODO: this only needs to happen when running alembic; it would be nice to check
create_test_mappers()
